#include <windows.h>
#include <ctype.h>
#include <string.h>
#include "modcliprc.h"

#define WM_TASKTRAY	(WM_USER+0x1000)
#define TASKTRAY_ID	(0)
#define IDM_QUIT	(0x100)

HINSTANCE hInst;
HMENU hMenu;
HWND hwndNextViewer = NULL;

static int
set_tasktray_icon(HWND hwnd)
{
    NOTIFYICONDATA nid;
    BOOL ret;

    memset(&nid, 0, sizeof (nid));
    nid.cbSize = sizeof (nid);
    nid.hWnd = hwnd;
    nid.uID = TASKTRAY_ID;
    nid.uFlags = NIF_ICON|NIF_MESSAGE|NIF_TIP;
    nid.uCallbackMessage = WM_TASKTRAY;
    nid.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_ICON));
    strcpy(nid.szTip, "clipboard modifier");
    ret = Shell_NotifyIcon(NIM_ADD, &nid);
    if (!ret)
	MessageBox(NULL, "cannot set task tray icon.", "Error", MB_OK);
    return !ret;
}

static void
remove_tasktray_icon(HWND hwnd)
{
    NOTIFYICONDATA nid;

    memset(&nid, 0, sizeof (nid));
    nid.cbSize = sizeof (nid);
    nid.hWnd = hwnd;
    nid.uID = TASKTRAY_ID;
    Shell_NotifyIcon(NIM_DELETE, &nid);
}

static int
create_tasktray_menu(HWND hwnd)
{
    hMenu = CreatePopupMenu();
    if (!hMenu)
	goto error;
    if (!AppendMenu(hMenu, MF_STRING, IDM_QUIT, "&Quit"))
	goto error;
    return 0;
error:
    MessageBox(NULL, "cannot create task tray menu.", "Error", MB_OK);
    return -1;
}

LRESULT
TaskTrayProc(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = 0;
    POINT point;
    RECT rect = { 0, 0, 0, 0 };
    HWND hwndShell;
    TPMPARAMS tpmp;

    switch (lParam) {
    case WM_RBUTTONDOWN:
	memset(&tpmp, 0, sizeof (tpmp));
	hwndShell = FindWindow("Shell_TrayWnd", NULL);
	if (hwndShell)
	    GetWindowRect(hwndShell, &rect);
	tpmp.cbSize = sizeof (tpmp);
	tpmp.rcExclude = rect;

	GetCursorPos(&point);
	SetForegroundWindow(hwnd);

	TrackPopupMenuEx(hMenu,
			 TPM_LEFTALIGN|TPM_LEFTBUTTON, point.x, point.y,
			 hwnd, &tpmp);
	break;
    }

    return ret;
}

HGLOBAL
modify_string(LPSTR lpstr)
{
    // XXX: hardwired. :-<
    BOOL isskipped = FALSE;
    BOOL isthereh = FALSE;
#define PREFIX "ttp://"
#define LEN_PREFIX (sizeof (PREFIX) -1)

    while (*lpstr && isspace(*lpstr)) {
	isskipped = TRUE;
	lpstr++;
    }

    if (*lpstr == 'h') {
	isthereh = TRUE;
	lpstr++;
    }
    if (!memcmp(lpstr, PREFIX, LEN_PREFIX)) {
	HGLOBAL hGlobal;
	LPSTR outstr, p;
	size_t len;
	BOOL isbs, isnewline, ismodified;

	len = strlen(lpstr) + 1 /* for 'h' */;
	hGlobal = GlobalAlloc(GMEM_MOVEABLE, len+1);
	outstr = GlobalLock(hGlobal);
	isbs = FALSE;
	ismodified = isskipped;
	*outstr++ = 'h';
	if (!isthereh) {
	    ismodified = TRUE;
	}
	while (*lpstr) {
	    isnewline = *lpstr == '\r' || *lpstr == '\n';
	    if (isspace(*lpstr) && !isnewline) {
		ismodified = TRUE;
		break;
	    }
	    if (isbs) {
		if (!isnewline)
		    *outstr++ = '\\';
		isbs = FALSE;
	    }
	    if (!isnewline) {
		if (*lpstr == '\\')
		    isbs = TRUE;
		else
		    *outstr++ = *lpstr;
	    } else
		ismodified = TRUE;
	    lpstr++;
	}
	*outstr = '\0';
	GlobalUnlock(hGlobal);
	if (ismodified)
	    return hGlobal;
	GlobalFree(hGlobal);
    }

    return NULL;
}

#define NumOf(array)	(sizeof (array) / sizeof (array[0]))
LRESULT
DrawClipboardProc(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = 0;
    int id;
    UINT formats[] = { CF_TEXT };
    HWND hwndowner;

    hwndowner = GetClipboardOwner();
    if (hwndowner == hwnd)
	return;

    id = GetPriorityClipboardFormat(formats, NumOf(formats));
    switch (id) {
    case CF_TEXT:
	if (OpenClipboard(hwnd)) {
	    HGLOBAL hGlobal, hGlbModified;
	    LPSTR str;
	    hGlobal = GetClipboardData(id);
	    str = GlobalLock(hGlobal);
	    hGlbModified = modify_string(str);
	    GlobalUnlock(hGlobal);
	    if (hGlbModified) {
		EmptyClipboard();
		SetClipboardData(CF_TEXT, hGlbModified);
	    }
	    CloseClipboard();
	}
	break;
    }

    return ret;
}

int
initialize(HWND hwnd)
{
    if (set_tasktray_icon(hwnd))
	return -1;
    if (create_tasktray_menu(hwnd))
	return -1;
    hwndNextViewer = SetClipboardViewer(hwnd);
    return 0;
}

void
cleanup(HWND hwnd)
{
    remove_tasktray_icon(hwnd);
    ChangeClipboardChain(hwnd, hwndNextViewer);
}

LRESULT CALLBACK
WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = 0;
    static PAINTSTRUCT ps;
    HDC hdc;

    switch (uMsg) {
    case WM_CREATE:
	if (initialize(hwnd))
	    PostQuitMessage(0);
	break;

    case WM_PAINT:
	hdc = BeginPaint(hwnd, &ps);
	EndPaint(hwnd, &ps);
	break;

    case WM_CLOSE:
	cleanup(hwnd);
	PostQuitMessage(0);
	break;

    case WM_DESTROY:
	cleanup(hwnd);
	break;

    case WM_TASKTRAY:
	ret = TaskTrayProc(hwnd, wParam, lParam);
	break;

    case WM_COMMAND:
	switch (wParam) {
	case IDM_QUIT:
	    PostMessage(hwnd, WM_CLOSE, 0, 0);
	    break;
	}
	break;

    case WM_CHANGECBCHAIN:
	if ((HWND)wParam == hwndNextViewer)
	    hwndNextViewer = (HWND)lParam;
	else if (hwndNextViewer != NULL)
	    SendMessage(hwndNextViewer, uMsg, wParam, lParam);
	break;

    case WM_DRAWCLIPBOARD:
	DrawClipboardProc(hwnd, wParam, lParam);
	SendMessage(hwndNextViewer, uMsg, wParam, lParam);
	break;

    default:
	ret = DefWindowProc(hwnd, uMsg, wParam, lParam);
    }

    return ret;
}

static int
register_class(void)
{
    WNDCLASS wc;

    memset(&wc, 0, sizeof (wc));
    wc.style = 0;
    wc.lpfnWndProc = &WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInst;
    wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_ICON));
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject(GRAY_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = "main window";
    if (!RegisterClass(&wc)) {
	MessageBox(NULL, "cannot register class", "Error", MB_OK);
	return -1;
    }
    return 0;
}

int
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
	int nCmdShow)
{
    HWND hwnd;
    MSG msg;

    hInst = hInstance;
    if (register_class())
	return 0;

    hwnd = CreateWindow("main window", "modclip",
			WS_OVERLAPPEDWINDOW,
			CW_USEDEFAULT, CW_USEDEFAULT,
			100, 100,
			NULL,
			NULL,
			hInstance,
			NULL);

    /* main loop */
    while (GetMessage(&msg, NULL, 0, 0)) {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }

    return 0;
}
